/**
 * @deps: "underscore"
 * @type {[function]}
 */
var _ = require("underscore"),
    request = require("request"),
    querystring = require("querystring"),
    crypto = require("crypto");

/**
 * @affect: use local log function
 * @developer: testing only, remove for production
 * @return {[type]} [description]
 */
var log = function() {
  console.log.apply(this, arguments);
};

/**
 * Global Settings
 * @type {Object}
 */
var settings = {
  url              : "graph.facebook.com",
  timeout          : 15000,
  appId            : "",
  appSecret        : "",
  signedRequestData: {},
  accessToken      : null
};

/**
 * @use: helpers
 * @purpose: helps prototype methods on new instances
 */
var base64UrlDecode,
    parseseSignedRequest,
    getAccessToken,
    api;

api = function(url, opts, callback) {

  var self = this;

/**
 * @affect: specify default options for requestjs
 * @type {Object}
 */
  var default_opts = {
    json: true,
    url: this._settings.url,
    timeout: this._settings.timeout,
    qs: {}
  };

/**
 * @affect: extend custom options
 */
  if (_.isObject(opts)) {

    opts = _.extend(default_opts, opts);

    if (_.has(opts, "https") && opts.https) {
      opts.url = "https://" + opts.url;
    }

    else {
      opts.url = "http://" + opts.url;
    }
  }

/**
 * @fail: use default opts
 * valid object wasn't there
 */
  else {
    opts.url = "http://" + opts.url;
    opts = default_opts;
  }

/**
 * @affect: use cache of access token
 */
  if (_.isObject(opts.qs) &&
      !(_.has(opts.qs, "access_token"))) {

    if (_.has(this._settings, "accessToken") &&
        _.isString(this._settings.accessToken)) {
      opts.qs.access_token = this._settings.accessToken;
    }

  }

/**
 * Correctly add slash after
 * domain name
 */
  opts.url += "/" + url;

/**
 * @affect: use requestjs to make a request
 * to the facebook api and overide its options with
 * the opts object given
 */
  request.get(opts, function(err, resp) {
    
    callback = _.bind(callback, self);

    if (err) {
      callback(err.message);
      return;
    }

    if (_.isObject(resp.body)) {

      if (_.has(resp.body, "error")) {
        callback(resp.body.error);
        return;
      }

    }

    callback(false, resp.body, resp);

  });

};

/**
 * @info: base64Decoder
 * @param  {[type]} str [description]
 * @return {[type]}     [description]
 */
base64UrlDecode = function (str) {

  var base64String, buffer;

  base64String = str.replace(/\-/g, '+').replace(/_/g, '/');
  buffer = new Buffer(base64String, 'base64');
  
  return buffer.toString('utf8');

};

/**
 * @info: parse Facebook signed request string
 * @param  {[type]} signedRequest [description]
 * @return {[type]}               [description]
 */
parseSignedRequest = function(signedRequest) {

  var appSecret = this._settings.appSecret || "",
      split,
      encodedSignature,
      encodedEnvelope,
      envelope,
      hmac,
      base64Digest,
      base64UrlDigest;

  if (!signedRequest) {
    signedRequest = this._settings.signedRequest;
  }

  if (!signedRequest || !appSecret) {
    return;
  }

  split = signedRequest.split('.');

  if (split.length !== 2) {
    return;
  }

  encodedSignature = split.shift();
  encodedEnvelope = split.shift();

  if(!encodedSignature || !encodedEnvelope) {
    return;
  }

  try {
    envelope = JSON.parse(base64UrlDecode(encodedEnvelope));
  } catch (e) {
    return;
  }

  if (!(envelope && _.has(envelope, 'algorithm') && envelope.algorithm.toUpperCase() === 'HMAC-SHA256')) {
    return;
  }

  hmac = crypto.createHmac('sha256', appSecret);
  hmac.update(encodedEnvelope);
  base64Digest = hmac.digest('base64');

  // remove Base64 padding
  base64UrlDigest = base64Digest.replace(/\={1,3}$/, '');

  // Replace illegal characters
  base64UrlDigest = base64UrlDigest.replace(/\+/g, '-').replace(/\//g, '_');

  if (base64UrlDigest !== encodedSignature) {
    return;
  }

  this._settings.signedRequestData = envelope;

  return envelope;

};

getAccessToken = function(arg1, arg2) {

  var code, callback;

  /**
   * @affect: ensure callback is function
   */
  if ("function" === typeof arg1) {
    callback = arg1;
  }

  else if ("function" === typeof arg2) {
    callback = arg2;
  }

  else {
    callback({
      code: 1,
      message: "Expected Auth Code"
    });
    return;
  }

/**
 * @affect: if arg1 is not a string
 * attempt to use stored signedRequestDat, else
 * attempt to directly parseSignedRequest and assign it as code
 * @fail: send error message
 */
  if (!_.isString(arg1)) {

  /**
   * @affect: determine wether we have
   * a cached signedRequestData.code
   */
    if (!this._settings.signedRequestData ||
        !_.isString(this._settings.signedRequestData.code)) {

    /**
     * @affect: we need to attempt to parse
     * the signed request now because all other
     * methods have been exhausted
     * @type {[type]}
     */
      var signedRequest = this.parseSignedRequest();

    /**
     * @affect: is the signed an object
     * and does it have a code
     */
      if (_.isObject(signedRequest) &&
          _.has(signedRequest, "code")) {
        code = signedRequest.code;
      }

    /**
     * @fail: we failed to obtain an
     * auth token, so error is the replied value
     */
      else {

        callback({
          code: 1,
          message: "Expected Auth Code"
        });

        return;

      }
    }
  /**
   * @success: use the cache signed request
   */
    else {
      code = this._settings.signedRequestData.code;
    }

  }
  /**
   * @affect: if arg1 is string use it
   */
  else {
    code = arg1;
  }

/**
 * @affect: Request map for facebook graph api
 * to call token and recieve string with access_token
 * and expires
 * @type {Object}
 */
  var map = {
    client_id: this._settings.appId,
    client_secret: this._settings.appSecret,
    code: code,
    redirect_uri: ""
  };

  /**
   * @affect: api request
   */
  this.api("oauth/access_token", {

    qs: map,
    https: true,
    json: true

  }, function(err, body) {

    var parsed_query;

    if (err) {
      callback(err);
      return;
    }

  /**
   * @affect: facebook gives us a query string
   * so we need to parse it here
   * @type {[type]}
   */
    parsed_query = querystring.parse(body);

    if (_.isObject(parsed_query) &&
        _.has(parsed_query, "access_token")) {

      this._settings.accessToken = parsed_query.access_token;
      callback(false, parsed_query.access_token);

    }

  /**
   * @fail: no access token was
   * sent reply with error
   */
    else {

      callback({
        code: 3,
        message: "No Access Token"
      });

      return;
    }
  
  });

};

/**
 * @info: Main Exports
 */
var Instance, MainExport;

/**
 * @info: Each instance created
 * @param {[type]} opt [description]
 */
Instance = function(opt) {
  this._settings = _.extend(_.clone(settings), opt);
};

/**
 * @info: prototype method for parse
 * @type {[type]}
 */
Instance.prototype.parseSignedRequest = parseSignedRequest;

/**
 * @info: Grab access with signed request code
 * @notice: will use arguments unless none specified otherwise it will check settings
 */
Instance.prototype.getAccessToken = getAccessToken;

/**
 * @affect: api method
 * @use: request api
 */
Instance.prototype.api = api;

/**
 * @info: Main exported object
 * @param {[type]} opt [description]
 */
MainExport = function(opt) {
  return new Instance(opt);
};

/**
 * @info: Overide global settings that
 * affect all new instances with the new
 * settings object passed here
 * @param {[type]} opt [description]
 */
MainExport.set = function(opt) {
  settings = _.extend(settings, opt);
  return settings;
};

/**
 * Export the application
 * @type {[type]}
 */
module.exports = MainExport;